#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <process.h>
#include <dos.h>
#include <errno.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <io.h>
#include "driver.h"

char * xmalloc(int size)
{
  char * res;
  extern char * malloc();

  res = malloc(size);
  if (res == NULL) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  return res;
}

int suffix(char * name, char * suff)
{
  int lname = strlen(name);
  int lsuff = strlen(suff);

  return lname >= lsuff && strcmp(name + lname - lsuff, suff) == 0;
}

int prefix(char * name, char * pref)
{
  return strncmp(name, pref, strlen(pref)) == 0;
}

char * change_suffix(char * name, char * suff)
{
  char * res, * dot;
  res = xmalloc(strlen(name) + strlen(suff) + 1);
  strcpy(res, name);
  dot = strrchr(res, '.');
  if (dot == NULL) {
    strcat(res, suff);
  } else {
    strcpy(dot, suff);
  }
  return res;
}

char * modulename(char * filename)
{
  char * p;
  /* Remove extension */
  p = filename + strlen(filename) - 1;
  while (*p != '.') p--;
  *p = 0;
  /* Find beginning of base name */
  while (p > filename) {
    p--;
    if (*p == '/' || *p == '\\' || *p == ':') return p+1;
  }
  return filename;
}

void format(char ** res, char * fmt, ...)
{
  va_list ap;
  char buffer[1024];
  
  va_start(ap, fmt);
  vsprintf(buffer, fmt, ap);
  *res = xmalloc(strlen(buffer) + 1);
  strcpy(*res, buffer);
  va_end(ap);
}

void reformat(char ** res, char * fmt, ...)
{
  va_list ap;
  char buffer[1024];
  int should_free;
  
  should_free = (**res != 0);
  va_start(ap, fmt);
  vsprintf(buffer, fmt, ap);
  if (should_free) free(*res);
  *res = xmalloc(strlen(buffer) + 1);
  strcpy(*res, buffer);
  va_end(ap);
}

void tempname(char * template)
{
  char * p;
  int n, i, j;

  for (p = template; *p != 'X'; p++) {
    if (*p == 0) return;
  }
  n = 0;
  do {
    for (i = n, j = 3; j >= 0; j--) {
      p[j] = '0' + i % 10;
      i = i / 10;
    }
    n++;
  } while (access(template, 0) == 0);
  return;
}

static int callcmd(char * cmd, char * args, int mode)
{
  int retcode;
#ifdef DEBUG
  fprintf(stderr, "\t%s %s\n", cmd, args);
#endif
  if (strlen(cmd) + 1 + strlen(args) < 128) {
    retcode = spawnlp(mode, cmd, cmd, args, NULL);
  } else {
    static char temparg[] = "@camlXXXX.$$$";
    char * tempfile = &temparg[1];
    int temp;
    tempname(tempfile);
    temp = open(tempfile, O_WRONLY | O_TRUNC | O_CREAT, S_IREAD | S_IWRITE);
    if (temp == -1) { perror(tempfile); exit(2); }
    write(temp, args, strlen(args));
    close(temp);
    retcode = spawnlp(P_WAIT, cmd, cmd, temparg, NULL);
    unlink(tempfile);
  }
  if (retcode == -1) {
    perror(cmd);
    return 2;
  }
  return retcode;
}

int execute(char * cmd, char * args)
{
  return callcmd(cmd, args, P_WAIT);
}

void terminate(char * cmd, char * args)
{
  exit(callcmd(cmd, args, P_OVERLAY));
}

int execute_with_output_to(char * cmd, char * args, char * outputname)
{
  int output, saved_stdout;
  int status;

  saved_stdout = dup(1);
  output = open(outputname, O_WRONLY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
  if (output == -1) return 2;
  dup2(output, 1);
  close(output);
  status = execute(cmd, args);
  dup2(saved_stdout, 1);
  close(saved_stdout);
  return status;
}

int append(char * srcname, char * dstname)
{
  int status, src, dst, n;
  char buffer[1024];
#ifdef DEBUG
  fprintf(stderr, "\tcat %s >> %s\n", srcname, dstname);
#endif
  src = open(srcname, O_RDONLY | O_BINARY);
  if (src == -1) { perror(srcname); return 1; }
  dst = open(dstname, O_WRONLY | O_APPEND | O_BINARY);
  if (dst == -1) { close(src); perror(dstname); return 1; }
  status = 0;
  while(1) {
    n = read(src, buffer, sizeof(buffer));
    if (n == -1) { perror(srcname); status = 1; break; }
    if (n == 0) break;
    if (write(dst, buffer, n) != n) { perror(dstname); status = 1; break; }
  }
  close(src);
  close(dst);
  return status;
}

